#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "file.h"

unsigned char *hex_str(char *str); //prototype

int main(int argc, char **argv)
{
	int addy, stop, i, n, length, no_bytes;
	unsigned long int hex;
	char *buffer = NULL;
	unsigned char *bytes = NULL;
	file_t rom, input;

	if((argc != 4) && (argc != 5))
	{
	 printf("usage: insert_asm rom.bin [asm.bin | $raw_hex] start_addy <stop_addy>\n");
	 return 0;
	}
	else if(argc == 5)
	{
	 sscanf(argv[4], "%x", &stop);
	 sscanf(argv[3], "%x", &addy);	//address in rom to insert file at

	 if(stop < addy)
	 {
	  printf("Stop address occurs before start address!\n");
	  return 1;
	 }
	}
	else
	{
	 stop = 0;
	 sscanf(argv[3], "%x", &addy);	//address in rom to insert file at
	}

	//read the input into a buffer so we can determine if it's raw hex or a file
	buffer = (char *)malloc(strlen(argv[2] + 1));
	strcpy(buffer, argv[2]);

	strcpy(rom.filename, argv[1]);	//load rom into buffers

	if(load_file(&rom))
	{
	 printf("A problem occured while opening the rom.\n");
	 return 1;
	}

	if(buffer[0] == '$')
	{
	 //buffer is raw hex
	 sscanf(buffer, "$%x", &hex);
	 length = strlen(buffer) - 1;

	 if((length % 2) == 0) //even number of characters, no need to add 0's to the front
	 {
	  bytes = hex_str(buffer + 1); //ignore the leading $
	  if(!bytes)
	  {
	   return 1;
	  }
	 }
	 else //odd number of characters, add a 0 in front to make an even length string
	 {
	  while(length % 2) //length must be a multiple of two
	  {
	   length++;
	   buffer[0] = '0'; //replace the '$' with a 0
	  }

	  bytes = hex_str(buffer);
	  if(!bytes)
	  {
	   return 1;
	  }
	 }

	 n = 0;
	 no_bytes = length / 2; //number of bytes in the hex string

	 for(i = 0; i < no_bytes; i++)
	 {
	  if(stop == (addy + i))
	  {
	   printf("Stop address reached. $%X of $%X bytes were written.\n", i, no_bytes);
	   printf("%s will be unaltered.\n", rom.filename);
	   return 1;
	  }
	  rom.ptr[addy + i] = bytes[i];
	 }

	 free(bytes);
	}
	else
	{
	 //buffer is a file to insert into rom.ext
	 strcpy(input.filename, argv[2]);

	 if(load_file(&input))
	 {
	  printf("A problem occured while opening the asm file.\n");
	  return 1;
	 }

	 for(i = 0; i < input.size; i++) //write contents of asm file
	 {
	  if(stop == (addy + i))
	  {
	   printf("Stop address reached. $%X of $%X bytes were written.\n", i, input.size);
	   printf("%s will be unaltered.\n", rom.filename);
	   return 1;
	  }
	  rom.ptr[addy + i] = input.ptr[i]; //copy one byte at a time
	 }
	} //else input == file

	write_file(rom); //overwrite the old rom
	printf("Finished inserting data.\n");

	return 0;
}

unsigned char *hex_str(char *str)
{
	int length, i = 0, j, n = 0;
	unsigned char *hex_str, temp;

	length = strlen(str);

	hex_str = (unsigned char *)malloc((length/2));
	if(!hex_str)
	{
	 hex_str = NULL;
	 printf("An error occured while converting the hexadecimal string.\n");
	 return hex_str;
	}

	while(i < length)
	{
	 hex_str[n] = 0;

	 for(j = 1; j >= 0; j--) //convert two chars to one byte
	 {
	  temp = str[i++];

	  if((temp >= '0') && (temp <= '9'))
	  {
	   temp -= 0x30; //convert from ascii to raw hex
	  }
	  else if((temp >= 'A') && (temp <= 'F'))
	  {
	   temp -= 0x37;
	  }
	  else if((temp >= 'a') && (temp <= 'f'))
	  {
	   temp -= 0x57;
	  }
	  else
	  {
	   hex_str = NULL;
	   printf("Encountered an invalid character in the hexadecimal string.\n");
	   return hex_str;
	  }

	  temp <<= (4 * j); //shift the raw hex to the correct nibble
	  hex_str[n] |= (temp & 0xff);
	 } //for(j...)

	 n++; //move to the next byte
	} //while(i < length)

	return hex_str; //return a pointer to the allocated string of bytes
}
